                 {* Charlie Wiederhold - Gooty Wed. 3-4 *}
UNIT TreeADT;
(* The operations in this unit are recursive where possible. *)

INTERFACE

(* Gain access to TreeElementType. *)
USES
  ListADT,
  TreeData;

TYPE
  ProcessType = procedure (Element : TreeElementType;
    VAR FileName : Text);
  TreePtrType = ^TreeNodeType;
  TreeNodeType = RECORD
    Info  : TreeElementType;    (* the user's data        *)
    Left  : TreePtrType;        (* pointer to left child  *)
    Right : TreePtrType         (* pointer to right child *)
  END;  (* TreeNodetype *)

  TreeType = TreePtrType;

(*********************************************************)

PROCEDURE CreateTree
  (VAR Tree : TreeType);
  (* Pre:  Tree has not been created or is empty. *)
  (* Post: Tree is empty.                         *)

(*********************************************************)

PROCEDURE RetrieveTreeElement
  (Tree        : TreeType;
   VAR Element : TreeElementType;
   VAR Found   : Boolean);
  (* Post:  Found is True if an element in the tree       *)
  (*        has the same key as Element; False otherwise. *)

(*********************************************************)

PROCEDURE ReplaceTreeElement
  (Tree        : TreeType;
   VAR Element : TreeElementType);
  (* Pre:  An element in the tree has the same key as Element. *)
  (* Post: Element replaces the element with the same key.     *)

(*********************************************************)

PROCEDURE InsertTreeElement
  (VAR Tree  : TreeType;
   Element   : TreeElementType);
  (* Pre:  There is not an element in the tree with the *)
  (*       same key as Element.                         *)
  (* Post: Element is in the tree.                      *)

(*********************************************************)

PROCEDURE DeleteTreeElement
  (VAR Tree : TreeType;
   Element  : TreeElementType);
  (* Pre:  Tree has been initialized.                       *)
  (*       An item with the same key as Element is in Tree. *)
  (* Post: No item with the same key as Element is in Tree. *)

(*********************************************************)

PROCEDURE DestroyTree
  (VAR Tree : TreeType);
  (* Pre:  Tree has been created                             *)
  (* Post: Tree has been destroyed to NIL                    *)

(*********************************************************)

IMPLEMENTATION

(*********************************************************)

PROCEDURE CreateTree
  (VAR Tree : TreeType);
BEGIN (* CreateTree *)
  Tree := NIL
END;  (* CreateTree *)

(*********************************************************)

PROCEDURE RetrieveTreeElement
  (Tree        : TreeType;
   VAR Element : TreeElementType;
   VAR Found   : Boolean);
  (* Pre:  Tree has been initialized.                       *)
  (* Post: If Compare(Element, SomeElement) is Equal,       *)
  (*         Found <-- True and Element <-- SomeElement;    *)
  (*         Else Found <-- False and Element is unchanged. *)
BEGIN  (* RetrieveElement *)
  IF Tree = NIL
    THEN
      Found := False
  ELSE
    CASE TreeCompare(Element, Tree^.Info) OF
      Less    : RetrieveTreeElement (Tree^.Left, Element, Found);
      Equal   : BEGIN
                  Found := True;
                  Element := Tree^.Info
                END;
      Greater : RetrieveTreeElement (Tree^.Right, Element, Found);
    END;  (* CASE *)
END;   (* RetrieveElement *)

(*********************************************************)

PROCEDURE ReplaceTreeElement
  (Tree        : TreeType;
   VAR Element : TreeElementType);
VAR
  Found : Boolean;
BEGIN  (* ReplaceTreeElement *)
  
  CASE TreeCompare(Element, Tree^.Info) OF
    Less    : ReplaceTreeElement (Tree^.Left, Element);
    Equal   : BEGIN
                Found := True;
                Tree^.Info := Element
              END;
    Greater : ReplaceTreeElement (Tree^.Right, Element);
  END;  (* CASE *)
END;   (* ReplaceTreeElement *)

(*********************************************************)

PROCEDURE InsertTreeElement
  (VAR Tree  : TreeType;
   Element   : TreeElementType);
  (* Pre:  Tree has been initialized.                        *)
  (*       No item in the tree has the same Key as Element.  *)
  (* Post: Element is in Tree; Tree is a binary search tree. *)
BEGIN (* InsertElement *)
  IF Tree = NIL 
    THEN  (* Base Case: Allocate new leaf Node. *)
      BEGIN
        New (Tree);  
        Tree^.Left  := NIL;
        Tree^.Right := NIL;
        Tree^.Info  := Element 
      END 
    ELSE 
      CASE TreeCompare (Element, Tree^.Info) OF
        Less    : (* General Case 1: Insert in left subtree. *)
                  InsertTreeElement (Tree^.Left, Element);
        Greater : (* General Case 2: Insert in right subtree. *)
                  InsertTreeElement (Tree^.Right, Element)
      END  (* CASE *)
END;  (* InsertElement *)

(*********************************************************)

PROCEDURE FindAndDeleteMax
  (VAR Tree    : TreeType;
   VAR Largest : TreeType);
  (* Pre:  Tree is not NIL.                   *)
  (* Post: Largest contains the item with the *)
  (*       maximum key value in Tree.         *)
  (*       Largest is deleted.                *)
BEGIN  (* FindAndDeleteMax *)
  IF Tree^.Right = NIL
     THEN  (* Base Case:  maximum found *)
       BEGIN
         Largest := Tree;
         Tree := Tree^.Left
       END
     ELSE  (* General Case: look in right subtree *)
       FindAndDeleteMax(Tree^.Right, Largest)
END;   (* FindAndDeleteMax *)

(*********************************************************)

PROCEDURE DeleteNode
  (VAR Tree : TreeType);
  (* Pre:  Tree is not NIL.               *)
  (* Post: Tree^.Info is not in the tree. *)
VAR
  TempPointer : TreeType;  (* used for Disposing *)
BEGIN  (* DeleteNode *)
  TempPointer := Tree;    (* save to Dispose later *)

  IF (Tree^.Left = NIL)
     THEN  (* Has no children or one child *)
       Tree := Tree^.Right
  ELSE IF (Tree^.Right = NIL)
     THEN  (* Has one child *)
       Tree := Tree^.Left
  ELSE (* Has two children *)
    BEGIN
      FindAndDeleteMax (Tree^.Left, TempPointer);
      DestroyList (Tree^.Info.BookList);
      Tree^.Info := TempPointer^.Info;
    END;
  Dispose (TempPointer)
END;   (* DeleteNode *)

(*********************************************************)

PROCEDURE DeleteTreeElement
  (VAR Tree : TreeType;
   Element  : TreeElementType);
  (* Pre:  Tree has been initialized.                       *)
  (*       An item with the same key as Element is in Tree. *)
  (* Post: No item with the same key as Element is in Tree. *)

BEGIN  (* DeleteElement *)
  CASE TreeCompare (Element, Tree^.Info) OF
    Less     : DeleteTreeElement (Tree^.Left, Element);
    Greater  : DeleteTreeElement (Tree^.Right, Element);
    Equal    : DeleteNode (Tree)
  END  (* CASE *)
END;   (* DeleteElement *)

(*********************************************************)

PROCEDURE DestroyTree
  (VAR Tree : TreeType);
  (* Pre:  Tree has been created                             *)
  (* Post: Tree has been destroyed to NIL                    *)

BEGIN (* DesrtroyTree *)
  WHILE Tree <> NIL DO DeleteNode (Tree);
END; (* DestroyTree *)

(*********************************************************)

END. (* TreeADT *)